// https://ifmagic.io/legal/eula-product-software

#include "NexusBPLibrary.h"

#include <string>
#include <iostream>
#include <sstream>
#include <stdio.h>

#include "SerialPort.h"
#include "TypeAbbreviations.h"

// TODO: convert to UStruct and broader Hardware UStruct
struct Port
{
	int module;
	FString data;
};
Port ports[8] = {{0, "0"}, {0, "0"}, {0, "0"}, {0, "0"}, {0, "0"}, {0, "0"}, {0, "0"}, {0, "0"}};
int portCount = 8;

// global data stream
std::string _stream;

// stream
// connect to a stream
bool UNexus::StreamConnect(FString portNumber, bool wireless)
{
	_stream = "";

	// create path name
	const char devicePathStr[] = "/dev/cu.SLAB_USBtoUART";
	const char devicePathStrWireless[] = "/dev/cu.Device";

	const int baudRate = 115200;

	// connect to port
	int sfd = 0;

	if (wireless)
	{
		sfd = openAndConfigureSerialPort(devicePathStrWireless, baudRate);
	}
	else
	{
		sfd = openAndConfigureSerialPort(devicePathStr, baudRate);
	}

	// see if port was opened
	if (sfd < 0)
	{
		if (sfd == -1)
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Orange, TEXT("Device Not Found"));
		}
		else
		{
			GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Orange, TEXT("Error Setting Port Attributes"));
		}
	}

	GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Blue, TEXT("Stream Started"));

	return true;
}

// disconnect from a stream
void UNexus::StreamDisconnect()
{
	closeSerialPort();
	GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Blue, TEXT("Stream Ended"));
}

// parse stream
void UNexus::StreamDimension(FString &stream)
{
	// file buffer variables
	char szBuff[1024];

	// string conversion variables
	std::string s;
	std::string fData;
	std::string data;

	ssize_t amount = readSerialData(szBuff, sizeof(szBuff)); // read data
	if (amount != 0)
	{
		// convert data to string
		s = convertToString(szBuff, sizeof(szBuff));

		// split new line
		std::istringstream iss(s);
		getline(iss, fData);

		// formatting
		int semiColonCount = std::count(fData.begin(), fData.end(), ';');
		if (semiColonCount >= (portCount + 1))
		{
			stream = UTF8_TO_TCHAR(fData.c_str());
			data = fData;
		}

		// split stream into ports
		for (int i = 0; i < portCount; i++)
		{

			// GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Green, (TEXT("%s"), UTF8_TO_TCHAR(data.c_str())));

			try
			{
				ports[i].module = atoi(data.substr(0, data.find(":")).c_str());
				// GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Blue, FString::Printf(TEXT("%d"), ports[i].module));
			}
			catch (std::invalid_argument const &e)
			{
				std::string _test = e.what();
			}
			catch (std::out_of_range const &e)
			{
				std::string _test = e.what();
			}

			try
			{
				ports[i].data = UTF8_TO_TCHAR(data.substr(data.find(":") + 1, data.find(";") - data.find(":") - 1).c_str());
				// GEngine->AddOnScreenDebugMessage(-1, 5.f, FColor::Orange, ports[i].data);
			}
			catch (std::invalid_argument const &e)
			{
				std::string _test = e.what();
			}
			catch (std::out_of_range const &e)
			{
				std::string _test = e.what();
			}

			data = data.erase(0, (data.find(";") + 1));
		}
	}
}

// send output to stream
void UNexus::DeviceOutput(FString value)
{
	// convert from FString to string
	std::string val = std::string(TCHAR_TO_UTF8(*value));
	std::string nVal = val + "\n";

	// file buffer variables
	char szBuff[256] = {0};

	// pass string characters to buffer
	for (int i = 0; i < 40; i++)
	{
		szBuff[i] = nVal[i];
	}

	// send data
	writeSerialData(szBuff, sizeof(szBuff));
}

// hardware
// abstract hardware data
void UNexus::Hardware(const FString stream, FString &port1, FString &port2, FString &port3, FString &port4, FString &port5, FString &port6, FString &port7, FString &port8)
{

	// convert FString to string
	std::string s = std::string(TCHAR_TO_UTF8(*stream));

	// give access to individual port data
	port1 = ports[0].data;
	port2 = ports[1].data;
	port3 = ports[2].data;
	port4 = ports[3].data;
	port5 = ports[4].data;
	port6 = ports[5].data;
	port7 = ports[6].data;
	port8 = ports[7].data;
}

// modules

// orientation
void UNexus::Orientation(const FString stream, float &xAxis, float &yAxis, float &zAxis, FRotator &orientation)
{
	std::string _data = std::string(TCHAR_TO_UTF8(*stream));

	if (_data.length() > 10)
	{

		for (int i = 0; i < 10; i++)
		{
			_data = _data.erase(0, (_data.find(";") + 1));
		}

		std::string o = _data.substr(0, _data.find(";"));

		o = o.erase(0, (o.find(":") + 1)); // remove orientation mode

		// set xAxis
		std::string _xAxis = o.substr(0, o.find(":"));
		if (_xAxis.length() > 0)
		{

			try
			{
				xAxis = std::stof(_xAxis);
				orientation.Yaw = xAxis;
			}
			catch (std::invalid_argument const &e)
			{
				std::string _test = e.what();
			}
		}

		// set yAxis
		o = o.erase(0, (o.find(":") + 1));
		std::string _yAxis = o.substr(0, o.find(":"));
		if (_yAxis.length() > 0)
		{

			try
			{
				yAxis = std::stof(_yAxis);
				orientation.Roll = yAxis;
			}
			catch (std::invalid_argument const &e)
			{
				std::string _test = e.what();
			}
		}

		// set zAxis
		o = o.erase(0, (o.find(":") + 1));
		std::string _zAxis = o.substr(0, o.find(":"));
		if (_zAxis.length() > 0)
		{

			try
			{
				zAxis = std::stof(_zAxis);
				orientation.Pitch = zAxis;
			}
			catch (std::invalid_argument const &e)
			{
				std::string _test = e.what();
			}
		}
	}
}

// button
void UNexus::Button(const int port, bool &found, bool &state)
{
	int mN = 1;			 // module number
	int mI = 0;			 // module index number
	bool f = false;		 // check if found
	bool search = false; // TODO: optimize performing search on start

	// automatically or manually find port
	if (search == false)
	{
		if (port == 0)
		{
			// automatic
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i; // set port index
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		// manual
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1; // set port index
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			state = (bool)std::stoi(std::string(TCHAR_TO_UTF8(*ports[mI].data)));
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// color
void UNexus::Color(const int port, bool &found, FLinearColor &color, int &red, int &green, int &blue, int &alpha)
{
	int mN = 4;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));
			std::string a = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string r = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string g = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string b = (s.substr(0, s.find(";")).c_str());

			alpha = std::stoi(a);
			red = std::stoi(r);
			green = std::stoi(g);
			blue = std::stoi(b);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// dial
void UNexus::Dial(const int port, bool &found, int &raw, int &degree)
{
	int mN = 2;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			float _raw = std::stof(s);

			float _degree = ((_raw) / 4095) * (290) + 40;

			raw = std::stoi(s);
			degree = static_cast<int>(_degree);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// digital
void UNexus::Digital(const int port, bool &found, bool &state)
{
	int mN = 10;
	int mI = 0;
	bool f = false;
	bool search = false;

	// automatically or manually find port
	if (search == false)
	{
		if (port == 0)
		{
			// automatic
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i; // set port index
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		// manual
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1; // set port index
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			state = (bool)std::stoi(std::string(TCHAR_TO_UTF8(*ports[mI].data)));
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// distance
void UNexus::Distance(const int port, bool &found, int &millimeters, int &inches)
{
	int mN = 8;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			float _millimeters = std::stof(s);

			float _inches = (((_millimeters) / 25.4) * (10)) / 10;

			millimeters = std::stoi(s);
			inches = static_cast<int>(_inches);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// environment
void UNexus::Environment(const int port, bool &found, float &temperatureCelsius, float &temperatureFahrenheit, float &humidity, float &gas, float &pressure)
{
	int mN = 21;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));
			std::string tC = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string p = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string h = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string g = (s.substr(0, s.find(";")).c_str());
			s = s.erase(0, (s.find(",") + 1));

			temperatureCelsius = std::stof(tC);
			temperatureFahrenheit = (((9 / 5) * temperatureCelsius * 10) / 10) + 32;
			humidity = std::stof(h);
			gas = std::stof(g);
			pressure = std::stof(p);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// flex
void UNexus::Flex(const int port, bool &found, int &raw, int &bend)
{
	int mN = 19;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			float _raw = std::stof(s);

			float _bend = ((_raw) / 1000) * (100);

			raw = std::stoi(s);
			bend = static_cast<int>(_bend);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// force
void UNexus::Force(const int port, bool &found, int &raw, int &strength)
{
	int mN = 20;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			float _raw = std::stof(s);

			float _strength = ((_raw) / 4095) * (100);

			raw = std::stoi(s);
			strength = static_cast<int>(_strength);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// gesture
void UNexus::Gesture(const int port, bool &found, EGestureDirection &direction)
{
	int mN = 3;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			int _raw = std::stoi(s);

			switch (_raw)
			{
			case 0:
				direction = EGestureDirection::NONE;
				break;
			case 1:
				direction = EGestureDirection::DOWN;
				break;
			case 2:
				direction = EGestureDirection::UP;
				break;
			case 3:
				direction = EGestureDirection::RIGHT;
				break;
			case 4:
				direction = EGestureDirection::LEFT;
				break;
			default:
				direction = EGestureDirection::NONE;
				break;
			}
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// glow
void UNexus::Glow(const int port, bool &found, FLinearColor color, FString &output)
{
	int mN = 9;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		std::string _port = std::to_string(mI + 1);

		FLinearColor HSV = color.LinearRGBToHSV();
		std::string _hue = std::to_string((int)(((float)255 / (float)360) * HSV.R));
		std::string _saturation = std::to_string((int)(HSV.G * 255));
		std::string _value = std::to_string((int)(255 - (HSV.B * 255)));

		std::string val = "3," + _port + ",9,0,6,0," + _hue + "," + _saturation + ":1," + _value;

		output = UTF8_TO_TCHAR(val.c_str());
	}
}

// joystick
void UNexus::Joystick(const int port, bool &found, int &rawX, int &rawY, int &x, int &y)
{
	int mN = 6;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			std::string _xraw = (s.substr(0, s.find(",")).c_str());
			s = s.erase(0, (s.find(",") + 1));
			std::string _yraw = (s.substr(0, s.find(";")).c_str());

			float _x = std::stof(_xraw);
			float _y = std::stof(_yraw);

			float _xR = ((_x) / 4095) * (100);
			float _yR = ((_y) / 4095) * (100);

			rawX = std::stoi(_xraw);
			rawY = std::stoi(_yraw);

			x = static_cast<int>(_xR);
			y = static_cast<int>(_yR);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// light
void UNexus::Light(const int port, bool &found, int &raw, int &brightness)
{
	int mN = 11;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		// TODO: generic parse data function
		std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

		float _raw;
		float _brightness;

		try
		{
			_raw = std::stof(s);
			_brightness = ((_raw) / 4095) * (100);
			brightness = static_cast<int>(_brightness);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}

		try
		{
			raw = std::stoi(s);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// motion
void UNexus::Motion(const int port, bool &found, bool &movement)
{
	int mN = 18;		 // module number
	int mI = 0;			 // module index number
	bool f = false;		 // check if found
	bool search = false; // TODO: optimize performing search on start

	// automatically or manually find port
	if (search == false)
	{
		if (port == 0)
		{
			// automatic
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i; // set port index
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		// manual
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1; // set port index
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{

		try
		{
			movement = (bool)std::stoi(std::string(TCHAR_TO_UTF8(*ports[mI].data)));
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// move
void UNexus::Move(const int port, bool &found, int degree, FString &output)
{
	int mN = 14;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		std::string _port = std::to_string(mI + 1);
		std::string _degree = std::to_string(degree);

		std::string val = "3," + _port + ",14," + _degree;

		output = UTF8_TO_TCHAR(val.c_str());
	}
}

// proximity
void UNexus::Proximity(const int port, bool &found, int &raw, int &amount)
{
	int mN = 5;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		// TODO: generic parse data function
		std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

		float _raw;
		float _amount;

		try
		{
			_raw = std::stof(s);
			_amount = ((_raw) / 255) * (100);
			amount = static_cast<int>(_amount);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}

		try
		{
			raw = std::stoi(s);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// slider
void UNexus::Slider(const int port, bool &found, int &raw, int &position)
{
	int mN = 15;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		// TODO: generic parse data function
		std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

		float _raw;
		float _position;

		try
		{
			_raw = std::stof(s);
			_position = ((_raw) / 4095) * (100);
			position = static_cast<int>(_position);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}

		try
		{
			raw = std::stoi(s);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// sound
void UNexus::Sound(const int port, bool &found, int &frequency)
{
	int mN = 12;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		// TODO: generic parse data function
		std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

		float _raw;
		float _frequency;

		try
		{
			_raw = std::stof(s);
			_frequency = ((_raw) / 4095) * (100);
			frequency = static_cast<int>(_frequency);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}

// spin
// void UNexus::Spin(const int port, bool& found, int& rotation, ESpinDirection& direction) {
//
//}

// thermal
// void UNexus::Thermal(const int port, bool& found, TArray<float>& gridCelsius, TArray<float>& gridFahrenheit, float& averageTemperatureCelsius, float& averageTemperatureFahrenheit, float& centerTemperatureCelsius, float& centerTemperatureFahrenheit) {
//
//}

// tone
void UNexus::Tone(const int port, bool &found, int frequency, int time, FString &output)
{
	int mN = 17;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		std::string _port = std::to_string(mI + 1);
		std::string _frequency = std::to_string(frequency);
		std::string _time = std::to_string(time);

		std::string val = "3," + _port + ",17," + _frequency + "," + _time;

		output = UTF8_TO_TCHAR(val.c_str());
	}
}

// touch
void UNexus::Touch(const int port, bool &found, int &presence)
{
	int mN = 16;
	int mI = 0;
	bool f = false;
	bool search = false;

	if (search == false)
	{
		if (port == 0)
		{
			for (int i = 0; i < portCount; i++)
			{
				if (ports[i].module == mN)
				{
					f = true;
					search = true;

					mI = i;
					break;
				}
				else
				{
					f = false;
					search = true;
				}
			}
		}
		else
		{
			if (ports[port - 1].module == mN)
			{
				f = true;
				search = true;

				mI = port - 1;
			}
			else
			{
				f = false;
				search = true;
			}
		}
	}

	if (f)
	{
		try
		{
			// TODO: generic parse data function
			std::string s = std::string(TCHAR_TO_UTF8(*ports[mI].data));

			float _raw = std::stof(s);

			float _presence = ((_raw) / 90) * (100);

			presence = static_cast<int>(_presence);
		}
		catch (std::invalid_argument const &e)
		{
			std::string _test = e.what();
		}
	}
}
